# -*- coding: utf-8; -*-
#
# oedes - organic electronic device simulator
# Copyright (C) 2017 Marek Zdzislaw Szymanski (marek@marekszymanski.com)
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License, version 3,
# as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

from . import base
import scipy.constants
from oedes import bdf1adapt as _bdf1adapt
from oedes import solve as _solve
from oedes import SolverError
from . import std
import numpy as np
import sys


def LEC(mesh, zc=1, za=-1, **kwargs):
    "Return ready-to-evaluate LEC model"
    model = base.BaseModel()
    std.electronic_device(model, mesh, 'pn', **kwargs)
    std.add_ions(model, mesh, zc=zc, za=za)
    model.setUp()
    return model


def _find(model):
    eq = dict([(eq.prefix, eq) for eq in model.species])
    return eq['cation'], eq['anion']


def initial_salt(model, cinit, nc=1., na=1.):
    "Return initial state of device corresponding to uniformly distributed ions with salt concentration cinit"

    cation, anion = _find(model)
    assert nc > 0 and na > 0 and cation.z > 0 and anion.z < 0
    assert cation.z * nc == -anion.z * na
    x = np.zeros_like(model.X)
    x[cation.idx] = cinit * nc
    x[anion.idx] = cinit * na
    return x


def debye_length(c, epsi_r=3., T=300.):
    return np.sqrt(scipy.constants.epsilon_0 * epsi_r * scipy.constants.Boltzmann *
                   T / (2 * scipy.constants.elementary_charge**2 * c))


def total_ions(model, eq, x):
    assert not eq.bc, "Calculation valid only for no-flux boundary conditions"
    return np.sum((eq.mesh.cells['volume'] * x[eq.idx]))


class checker:
    "Special variant of solve which checks if ion concentration is conservation to required precision"

    # This may but should not be useful
    # Solver may not conserve total number of ions if Jacobian is almost
    # singular

    def __init__(self, model, xinit, rtol, atol):
        self.model = model
        self.cation, self.anion = _find(model)
        self.rtol = rtol
        self.atol = atol
        self.c_tot = total_ions(model, self.cation, xinit)
        self.a_tot = total_ions(model, self.anion, xinit)

    def solve(self, model, *args, **kwargs):
        assert model is self.model
        x = _solve(model, *args, **kwargs)
        if not np.allclose(total_ions(model, self.cation, x),
                           self.c_tot, atol=self.atol, rtol=self.rtol):
            raise SolverError()
        if not np.allclose(total_ions(model, self.anion, x),
                           self.a_tot, atol=self.atol, rtol=self.rtol):
            raise SolverError()
        return x


def bdf1adapt(model, x, params, t, t1, dt,
              ion_atol=0., ion_rtol=1e-6, **kwargs):
    "bdf1adapt which includes checking of ion conservation (which should be totally unnecessary)"

    solve_and_check = checker(model, x, rtol=ion_rtol, atol=ion_atol)
    return _bdf1adapt(model, x, params, t, t1, dt,
                      solve=solve_and_check.solve, **kwargs)


class monitor_species:
    "Simple monitor, which plots species in Jupyter"

    def __init__(self, model, plt, ylim_species=[1e10, 1e30]):
        self.model = model
        self.plt = plt
        self.ylim_species = ylim_species

    def __call__(self, t, x, xt, out):
        plt = self.plt
        for eq in self.model.species:
            plt.plot(eq.mesh.cells['center'] * 1e9,
                     out[eq.prefix + '.c'], label=eq.prefix)
        plt.yscale('log')
        plt.xlabel('Position [nm]')
        plt.ylabel('Concentration [$m^-3$]')
        plt.legend(loc=0)
        plt.title(str(t))
        plt.ylim(self.ylim_species)
        plt.show()


def monitor_time(t, x, xt, out):
    sys.stdout.write('\r%s' % t + ' ' * 20)


def equilibrium(model, params, cinit, t1, dt, monitor=lambda t, x,
                xt, out: None, nc=1., na=1., ion_atol=0., ion_rtol=1e-6, **kwargs):
    # Do not modify params, so create copy
    params = params.copy()
    params['electrode1.voltage'] = 0.
    params['electrode0.voltage'] = 0.
    xinit = initial_salt(model, cinit, nc=nc, na=na)
    for t, x, xt, outf in bdf1adapt(
            model, xinit, params, 0., t1, dt, ion_atol=ion_atol, ion_rtol=ion_rtol):
        monitor(t, x, xt, outf())
    return x


def species_v_D_limited_from_params(eq, vars):
    v, D = species_v_D_charged_from_params(eq, vars)
    return v * (1. - eq.mesh.faceaverage(vars['c'][eq.prefix + '.N0'])), D



# -*- coding: utf-8; -*-
#
# oedes - organic electronic device simulator
# Copyright (C) 2017 Marek Zdzislaw Szymanski (marek@marekszymanski.com)
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License, version 3,
# as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

from .base import *
from .simple import *
from .boundary import *


class species_from_dos(object):
    # This is currently badly designed and will be refactored later

    def __init__(self, dos_class):
        self.dos_class = dos_class

    def _make_dos(self, eq, vars):
        params = vars['params']
        base = self.dos_class(params[eq.prefix + '.N0'], params['T'])
        return MO(base, params[eq.prefix + '.level'], eq.z)

    def v_D(self, eq, vars):
        v, D0 = species_v_D_charged_from_params(eq, vars)
        dos = self._make_dos(eq, vars)
        return v, eq.mesh.faceaverage(
            vars['mu'][eq.prefix] * dos.g(vars['c'][eq.prefix]))

    def bc(self, eq, facefluxes, vars):
        dos = self._make_dos(eq, vars)
        return bc_dirichlet(eq, vars['c'][eq.prefix], lambda u: dos.concentration(
            vars['params']['%s.workfunction' % u]))


class DOS(object):

    def Ef(self, eq, vars):
        "Return band-referenced Fermi level from concentration"
        raise NotImplementedError()

    def c(self, eq, Ef, vars):
        "Return concentration from band-referenced Fermi level"
        raise NotImplementedError()

    def D(self, eq, c, Ef, vars):
        "Return diffusion coefficient"
        raise NotImplementedError()

    def QuasiFermiLevel(self, eq, vars):
        assert eq.z in [1, -1]
        Eband = -vars['potential'] - vars['params'][eq.prefix + '.level']
        return Eband - eq.z * self.Ef(eq, vars)

    def concentration(self, eq, vars, idx, imref):
        assert eq.z in [1, -1]
        Eband = -vars['potential'][idx] - vars['params'][eq.prefix + '.level']
        return self.c(eq, -(imref - Eband) / eq.z, vars)


class BoltzmannDOS(DOS):
    c_eps = 1e-30
    c_limit = True
    Ef_max = None

    def Ef(self, eq, vars):
        c = vars['c'][eq.prefix]
        c = where(c > self.c_eps, c, self.c_eps)  # avoid NaN
        N0 = vars['params'][eq.prefix + '.N0']
        # c=where(c<N0,c,N0)
        return vars['Vt'] * np.log(c / N0)

    def c(self, eq, Ef, vars):
        if self.Ef_max is not None:
            Ef = where(Ef < self.Ef_max, Ef, self.Ef_max)
        N0 = vars['params'][eq.prefix + '.N0']
        v = Ef / vars['Vt']
        if self.c_limit:
            v = where(v < 0., v, 0.)
        c = np.exp(v) * N0
        return c

    def D(self, eq, c, Ef, vars):
        return vars['Vt']

    def v_D(self, eq, vars):
        return species_v_D_charged_from_params(eq, vars)



# -*- coding: utf-8; -*-
#
# oedes - organic electronic device simulator
# Copyright (C) 2017 Marek Zdzislaw Szymanski (marek@marekszymanski.com)
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License, version 3,
# as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

from .base import *
from oedes import functions


class BulkSource(object):
    def setUp(self, model):
        pass

    def evalute(self, vars):
        raise NotImplementedError()


class Interface(object):

    def __init__(self, equation, other_equation, name):
        self.equation, self.other_equation, self.name = equation, other_equation, name

    def setUp(self, model):
        self.mesh = self.equation.mesh
        self.other_mesh = self.other_equation.mesh
        self.boundary, self.other_boundary = self.mesh.getSharedBoundary(
            self.other_mesh, self.name)
        self.bidx, self.other_bidx = self.boundary[
            'bidx'], self.other_boundary['bidx']
        self.idx, self.other_idx = self.mesh.boundary.idx[
            self.bidx], self.other_mesh.boundary.idx[self.other_bidx]

    def evaluate(self, model, vars, vars_eq, other_vars_eq):
        raise NotImplementedError()

    def run(self, model, vars):
        eqvars = vars['eqvars']
        vars_eq = eqvars[id(self.equation)]
        other_vars_eq = eqvars[id(self.other_equation)]
        FdS, other_FdS, sources, other_sources = self.evaluate(
            model, vars, vars_eq, other_vars_eq)

        def add(dst, terms, bidx):
            for p, f in terms:
                dst[p].append((bidx, f))
        add(vars_eq['boundary_FdS'], FdS, self.bidx)
        add(other_vars_eq['boundary_FdS'], other_FdS, self.other_bidx)
        add(vars_eq['boundary_sources'], sources, self.bidx)
        add(other_vars_eq['boundary_sources'], other_sources, self.other_bidx)


class _Recombination(BulkSource):
    "Langevinian recombination term, to be used"

    def __init__(self, mesh, electron_prefix='electron',
                 hole_prefix='hole', output_name='R'):
        self.mesh = mesh
        self.electron = electron_prefix
        self.hole = hole_prefix
        self.output_name = output_name

    @property
    def minus(self):
        return [self.electron, self.hole]

    @property
    def plus(self):
        return []

    def evaluate(self, vars):
        R = self.evaluate_recombination(vars)
        full_output = vars['full_output']
        if full_output is not None:
            assert self.output_name not in full_output
            full_output[self.output_name] = R
        return R


class LangevinRecombination(_Recombination):
    def evaluate_recombination(self, vars):
        mu = vars['mu']
        c = vars['c']
        params = vars['params']
        return functions.LangevinRecombination(mu[self.electron], mu[self.hole], c[self.electron], c[
            self.hole], vars['epsilon'], npi=vars['params']['npi'])


class DirectRecombination(_Recombination):
    def __init__(self, *args, **kwargs):
        param_name = kwargs.pop('param_name', 'beta')
        super(DirectRecombination, self).__init__(*args, **kwargs)
        self.param_name = param_name

    def evaluate_recombination(self, vars):
        mu = vars['mu']
        c = vars['c']
        return vars['params'][self.param_name] * \
            (c[self.electron] * c[self.hole] - vars['params']['npi'])


class DirectGeneration(BulkSource):
    def __init__(self, eqs, function, prefix='absorption'):
        assert sum(eq.z for eq in eqs) == 0
        self.function = function
        self.eqs = eqs
        self.prefix = prefix
        self.G = self.function(self.eqs[0].mesh.cells['center'])

    @property
    def plus(self):
        return [eq.prefix for eq in self.eqs]

    @property
    def minus(self):
        return []

    def evaluate(self, vars):
        g = self.G * vars['params'][self.prefix + '.I']
        if vars['full_output'] is not None:
            vars['full_output'][self.prefix + '.G'] = g
        return g


class SimpleGenerationTerm(BulkSource):
    def __init__(self, eq, function, prefix='absorption'):
        self.function = function
        self.eq = eq
        self.prefix = self.eq.prefix + '.' + prefix
        self.G = self.function(self.eq.mesh.cells['center'])

    @property
    def plus(self):
        return [self.eq.prefix]

    @property
    def minus(self):
        return []

    def evaluate(self, vars):
        g = self.G * vars['params'][self.prefix + '.I']
        if vars['full_output'] is not None:
            vars['full_output'][self.prefix + '.G'] = g
        return g


class SimpleDecayTerm(BulkSource):
    def __init__(self, eq, prefix='decay'):
        assert eq.z == 0
        self.eq = eq
        self.prefix = self.eq.prefix + '.' + prefix

    @property
    def minus(self):
        return [self.eq.prefix]

    @property
    def plus(self):
        return []

    def evaluate(self, vars):
        return vars['c'][self.eq.prefix] * vars['params'][self.prefix]


class OnsagerBraunRecombinationDissociationTerm(BulkSource):
    b_eps = 1e-10
    b_max = 100

    def __init__(self, eq, electron_eq, hole_eq,
                 prefix='dissociation', binding_energy_param=False):
        assert eq.z == electron_eq.z + hole_eq.z, 'inconsistent signs'
        self.eq = eq
        self.electron_eq = electron_eq
        self.hole_eq = hole_eq
        self.prefix = self.eq.prefix + '.' + prefix
        self.binding_energy_param = binding_energy_param

    @property
    def plus(self):
        return [self.eq.prefix]

    @property
    def minus(self):
        return [self.electron_eq.prefix, self.hole_eq.prefix]

    def evaluate(self, vars):
        mu = vars['mu']
        c = vars['c']
        params = vars['params']
        full_output = vars['full_output']
        if self.binding_energy_param:
            Eb = params[self.eq.prefix + '.binding_energy']
            a = scipy.constants.elementary_charge / \
                (Eb * 4 * np.pi * vars['epsilon'])
        else:
            a = params[self.eq.prefix + '.distance']
            Eb = scipy.constants.elementary_charge / \
                (4 * np.pi * vars['epsilon'] * a)
        u = 3. / (4. * np.pi * a**3)  # m^-3
        v = np.exp(-Eb / vars['Vt'])  # 1
        b = (scipy.constants.elementary_charge / (8 * np.pi)) * \
            vars['Ecellm'] / (vars['epsilon'] * vars['Vt']**2)  # 1
        # b=0.
        t = functions.OnsagerFunction(
            where(
                b < self.b_max,
                b,
                self.b_max) +
            self.b_eps)  # 1
        gamma = scipy.constants.elementary_charge * \
            (mu[self.electron_eq.prefix] + mu[self.hole_eq.prefix]) / \
            vars['epsilon']  # m^3 /s
        # print('\ngamma=%s,\nc=%s,u,v,t=%s\n,%s\n,%s\n'%(gamma.value,c[self.eq.prefix].value,u,v,t.value))
        r = gamma * (c[self.electron_eq.prefix] *
                     c[self.hole_eq.prefix] - params['npi'])  # 1/(m^3 s)
        d = gamma * c[self.eq.prefix] * u * v * t
        if full_output is not None:
            full_output[self.prefix + '.recombination'] = r
            full_output[self.prefix + '.dissociation'] = d
        return r - d


class TrapSource(BulkSource):
    "Source term fro conduction level to trap level"

    def __init__(self, transport_eq, trap_eq, fill_transport=True,
                 fill_trap=True, rrate_param=False):
        assert transport_eq.mesh is trap_eq.mesh, "the same mesh must be used for conduction and trap level"
        assert transport_eq.z == trap_eq.z, "inconsistent signs"

        self.transport_eq = transport_eq
        self.trap_eq = trap_eq
        self.fill_transport = fill_transport
        self.fill_trap = fill_trap
        self.rrate_param = rrate_param

    @property
    def minus(self):
        return [self.transport_eq.prefix]

    @property
    def plus(self):
        return [self.trap_eq.prefix]

    def evaluate(self, vars):
        params = vars['params']
        c = vars['c']
        transportlevel = params[self.transport_eq.prefix + '.level']
        trate = params[self.trap_eq.prefix + '.trate']
        if self.rrate_param:
            rrate = params[self.trap_eq.prefix + '.rrate']
        else:
            traplevel = params[self.trap_eq.prefix + '.level']
            depth = (transportlevel - traplevel) * self.transport_eq.z
            rrate = trate * exp(-depth / vars['Vt'])

        transportN0 = params[self.transport_eq.prefix + '.N0']
        trapN0 = params[self.trap_eq.prefix + '.N0']
        ctransport = c[self.transport_eq.prefix]
        ctrap = c[self.trap_eq.prefix]
        if self.fill_trap:
            a = trate * (trapN0 - ctrap)
        else:
            a = trate * trapN0
        if self.fill_transport:
            b = rrate * (transportN0 - ctransport)
        else:
            b = rrate * transportN0
        return a * ctransport - b * ctrap


class SRHRecombination(object):
    def __init__(self, transport_eq, trap_eq):
        assert transport_eq.z == -trap_eq.z, 'inconsistent signs'
        self.transport_eq = transport_eq
        self.trap_eq = trap_eq

    @property
    def plus(self):
        return []

    @property
    def minus(self):
        return [self.transport_eq.prefix, self.trap_eq.prefix]

    def evaluate(self, vars):
        mu = vars['mu']
        c = vars['c']
        return scipy.constants.elementary_charge * \
            mu[self.transport_eq.prefix] * c[self.transport_eq.prefix] * \
            c[self.trap_eq.prefix] / vars['epsilon']


class SRH(BulkSource):
    "SRH (steady-state) term"

    def __init__(self, electron_eq, hole_eq, prefix='srh'):
        assert electron_eq.mesh is hole_eq.mesh, "the same mesh must be used for conduction and trap level"
        assert hole_eq.z == 1 and electron_eq.z == -1
        self.electron_eq = electron_eq
        self.hole_eq = hole_eq
        self.prefix = prefix

    @property
    def minus(self):
        return [self.electron_eq.prefix, self.hole_eq.prefix]

    @property
    def plus(self):
        return []

    def evaluate(self, vars):
        params, Vt, c = vars['params'], vars['Vt'], vars['c']
        Et = params[self.prefix + '.level']
        ni = params[self.electron_eq.prefix + '.N0'] * \
            exp((params[self.electron_eq.prefix + '.level'] - Et) / Vt)
        pi = params[self.hole_eq.prefix + '.N0'] * \
            exp((Et - params[self.hole_eq.prefix + '.level']) / Vt)
        Cn = params[self.electron_eq.prefix + '.' + self.prefix + '.trate']
        Cp = params[self.hole_eq.prefix + '.' + self.prefix + '.trate']
        n = c[self.electron_eq.prefix]
        p = c[self.hole_eq.prefix]
        g = Cn * Cp * params[self.prefix + '.N0'] * \
            (n * p - ni * pi) / (Cn * (n + ni) + Cp * (p + pi))
        if vars['full_output'] is not None:
            vars['full_output'][self.prefix + '.' +
                                self.hole_eq.prefix + '.G'] = g
            vars['full_output'][self.prefix + '.' +
                                self.electron_eq.prefix + '.G'] = g
        return g



# -*- coding: utf-8; -*-
#
# oedes - organic electronic device simulator
# Copyright (C) 2017 Marek Zdzislaw Szymanski (marek@marekszymanski.com)
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License, version 3,
# as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import tmm
import scipy.interpolate
import numpy as np


class SimpleLayer:
    def __init__(self, thickness, n):
        self.d = thickness
        self._n = n

    def n(self, wavelength):
        if callable(self._n):
            return self._n(wavelength)
        else:
            return self._n


class SimpleAbsorption:
    def __init__(self, mesh, n, layers_before, layers_after, wavelength):
        wrap = SimpleLayer(mesh.length, n)
        layers = layers_before + [wrap] + layers_after
        d_list = ['inf'] + [u.d for u in layers] + ['inf']
        n_list = [1.] + [u.n(wavelength) for u in layers] + [1.]
        out = tmm.coh_tmm(
            pol='s',
            n_list=n_list,
            d_list=d_list,
            th_0=0.,
            lam_vac=wavelength)
        i = len(layers_before) + 1
        x = mesh.cells['center']
        absor = np.asarray(
            [tmm.position_resolved(i, x_, out)['absor'] for x_ in x])
        self.normalization = np.amax(absor)
        self.a = scipy.interpolate.interp1d(x, absor)

    def normalized(self, x):
        return self.a(x) / self.normalization



# -*- coding: utf-8; -*-
#
# oedes - organic electronic device simulator
# Copyright (C) 2017 Marek Zdzislaw Szymanski (marek@marekszymanski.com)
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License, version 3,
# as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

from .base import *
from .dos import *
from .boundary import *
from .source import *
from oedes.fvm import mesh1d


def add_transport(model, mesh, z, prefix, traps=[], dos_class=BoltzmannDOS):
    """
    Adds transport equation to the model.
    Also adds trap levels and source terms transport<->trap level.
    """

    eq = TransportCharged(mesh, prefix, z=z)
    model.species.append(eq)
    if dos_class is not None:
        model.species_dos[eq.prefix] = dos_class()
    eq.bc = [FermiLevelEqualElectrode(boundary)
             for boundary in mesh.boundaries]
    model.species_v_D[eq.prefix] = species_v_D_charged_from_params
    for name in traps:
        teq = TransportCharged(mesh, eq.prefix + '.' + name, z=z)
        model.species.append(teq)
        model.species_v_D[teq.prefix] = species_v_D_not_transported
        #teq.bc = [Zero(boundary) for boundary in mesh.boundaries]
        model.sources.append(TrapSource(eq, teq))
    return eq


def electronic_device(model, mesh, polarity, ntraps=[], ptraps=[], **kwargs):
    """
    Creates simple electronic device:
    - if Poisson's equation is not present, create default with voltage
    - creates 'hole' if 'p' polarity requested, optionally with corresponding traps
    - creates 'electron' if 'n' polarity requested, optionally with corresponding traps
    - if both polarities are present, adds recombination term
    """
    if model.poisson is None:
        model.poisson = Poisson(mesh)
        model.poisson.bc = [AppliedVoltage(boundary)
                            for boundary in mesh.boundaries]
    assert polarity in ['n', 'p', 'np', 'pn', ''], "invalid polarity"
    electron, hole = None, None
    if 'n' in polarity:
        electron = add_transport(
            model, mesh, -1, 'electron', traps=ntraps, **kwargs)
    if 'p' in polarity:
        hole = add_transport(model, mesh, +1, 'hole', traps=ptraps, **kwargs)
    if electron is not None and hole is not None:
        r = LangevinRecombination(
            mesh, electron_prefix=electron.prefix, hole_prefix=hole.prefix)
        model.sources.append(r)


def add_ions(model, mesh, zc=1, za=-1):
    assert zc > 0 and za < 0, "follow standard convention"
    cation = TransportCharged(mesh, 'cation', zc)
    anion = TransportCharged(mesh, 'anion', za)
    model.species.extend([cation, anion])
    for prefix in ['cation', 'anion']:
        model.species_v_D[prefix] = species_v_D_charged_from_params
    return cation, anion


def holeonly(L, traps=[]):
    model = BaseModel()
    electronic_device(model, mesh1d(L), 'p', ptraps=traps)
    model.setUp()
    return model


def electrononly(L, traps=[]):
    model = BaseModel()
    electronic_device(model, mesh1d(L), 'n', ntraps=traps)
    model.setUp()
    return model



# -*- coding: utf-8; -*-
#
# oedes - organic electronic device simulator
# Copyright (C) 2017 Marek Zdzislaw Szymanski (marek@marekszymanski.com)
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License, version 3,
# as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

# Basic equations
#
# Using vertex centered finite volume method to discretize the equations.
# $\frac{du}{dt} = S + \nabla \cdot F $
# by discretized applying Gauss theorem
# $\int_V \frac{du}{dt}dV = \int_V S dV + \int_\Omega F d\Omega$
# where $V$ denotes cell volume, and $\Omega$ denotes cell boundary.
#
# In discretized geometry
# $V \frac{dc_i}{dt} = V s_i + \sum_k A_{ik} f_{k}$
# This is conveniently expressed by Equation.conservation
# $\frac{d\mathbf{c}}{dt} - \mathbf{s} - V^-1 \mathbf{A} \mathbf{f}=F_{internal}$
# where $V^-1$ is diagonal matrix of cell volumes. A is called fluxsum in the code.
#
# At boundary, values calculated by above equation are replaced by
# $b(V^-1 \mathbf{A} \mathbf{f},...)=F_{boundary}$
#
# Final residuals are given as $F_{internal}$ for internal points, and
# $F_{boundary}$ for boundary cells.

from oedes import ad
import numpy as np


class ConvergenceTest(object):
    pass


class NormConvergenceTest(ConvergenceTest):

    def __init__(self, stol=1e-6, atol=0.):
        self.stol = stol
        self.atol = atol

    def testEquation(self, eq, F, x, dx, params):
        F = F[eq.idx]
        x = x[eq.idx]
        dx = dx[eq.idx]
        if np.linalg.norm(dx) < self.stol * np.linalg.norm(x):
            return True
        if np.linalg.norm(F) < self.atol * np.sqrt(len(F)):
            return True
        return False

    def testBoundary(self, eq, bc, F, x, dx, params):
        raise NotImplementedError()


class ElementwiseConvergenceTest(ConvergenceTest):

    def __init__(self, rtol=1e-6, atol=0., drtol=1e-6, datol=0.):
        self.rtol = rtol
        self.atol = atol

    def _testix(self, eq, ix, F, x, dx, report, prefix):
        i = eq.idx[ix]
        if len(i) == 0:
            return True
        F = F[i]
        x = x[i]
        dx = dx[i]
        converged_abs = np.abs(F) < self.atol
        converged_rel = np.abs(dx) < self.rtol * np.abs(x)
        converged = np.logical_or(converged_abs, converged_rel).all()

        def amax(v):
            if len(v) == 0:
                return 0.
            else:
                return np.amax(v)
        with np.errstate(divide='ignore', invalid='ignore'):
            report[prefix] = dict(converged=converged,
                                  atol=self.atol,
                                  rtol=self.rtol,
                                  maxabs=amax(np.abs(F)),
                                  maxabs_nc=amax(
                                      np.abs(F[np.logical_not(converged_rel)])),
                                  maxrel=amax(np.abs(dx / x)),
                                  maxrel_nc=amax(np.abs((dx / x)[np.logical_not(converged_abs)])))
        return converged

    def testEquation(self, eq, F, x, dx, params, report):
        ix_all = np.hstack(
            [eq.mesh.internal.idx, eq.mesh.boundary.idx[eq.bc_free_dof]])
        return self._testix(eq, ix_all, F, x, dx, report, eq.prefix)

    def testBoundary(self, eq, bc, F, x, dx, params, report):
        return self._testix(eq, eq.mesh.boundary.idx[
                            bc.bidx], F, x, dx, report, eq.prefix + '.' + bc.name)


class BoundaryEquation(object):
    idx = None
    mesh = None
    bidx = None
    conservation_to = None

    convergenceTest = None

    def setUp(self, eq):
        assert self.idx is None and self.mesh is None, 'object already setUp'
        self.idx = eq.idx
        self.mesh = eq.mesh
        self.conservation_to = self.idx[
            self.mesh.boundary.idx[self.bidx]].copy()

    def conservation(self, eq, FdS, celltransient=0., cellsource=0.):
        if not ad.isscalar(celltransient):
            celltransient = celltransient[self.bidx]
        if not ad.isscalar(cellsource):
            cellsource = cellsource[self.bidx]
        volume = eq.boundary.cells['volume'][self.bidx]
        return -FdS + volume * (celltransient - cellsource)


class CellEquation(object):
    convergenceTest = None
    idx = None
    bc = None
    bc_free_dof = None
    bc_conservation_labels = None

    """
    Basic equation defining one degree of freedom per cell, solved using vertex centered finite volume

    Properties
    ----------

    transientvar : float
        one if pseudotransient techniques should be applied to this equation, zero if equation is time-independent
    idx : array of int
        indices of degrees of freedom in solution vector corresponding to this equation
    convergenceTest : ConvergenceTest
        convergence criteria for this equation
    prefix : str
        prefix of this equation used for parameters and output
    """

    def __init__(self, mesh, name, boundary=None):
        """Abstract block of equation, which introduces a field into equation system on :mesh:
        Keeps track of:
        indices of belonging to the field :idx:, x[idx[i]] is field in cell i
        :name:, which currently is the same as prefix
        :boundary: cells
        """
        self.mesh = mesh
        self.name = name
        self.bc = []

    @property
    def prefix(self):
        return self.name

    def residuals(self, part, facefluxes, celltransient=0., cellsource=0.):
        """
        Calculate FVM residuals

        Parameters:
        -----------
        facefluxes: vector or None
            fluxes per unit of surface or None
        celltransient: float or vector
            transient term per unit of volume
        cellsource: float or vector
            source term per unit of volume


        Returns: tuple (residuals,FdS)
        --------
        residuals:
            residuals for each cell, per unit of volume
        FdS:
            integral FdS per cell
        """

        assert part is self.mesh.internal or part is self.mesh.boundary
        idx = part.idx
        if facefluxes is None:
            FdS = 0.
        else:
            FdS = ad.dot(part.fluxsum, facefluxes)
        if not ad.isscalar(celltransient):
            celltransient = celltransient[idx]
        if not ad.isscalar(cellsource):
            cellsource = cellsource[idx]
        return self.idx[idx], -FdS / \
            part.cells['volume'] + celltransient - cellsource

    @property
    def transientvar(self):
        return 1.

    def testConverged(self, F, x, dx, params, report):
        tests = [self.convergenceTest.testEquation(self, F, x, dx, params, report)] + [
            bc.convergenceTest.testBoundary(self, bc, F, x, dx, params, report) for bc in self.bc]
        return all(tests)



# -*- coding: utf-8; -*-
#
# oedes - organic electronic device simulator
# Copyright (C) 2017 Marek Zdzislaw Szymanski (marek@marekszymanski.com)
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License, version 3,
# as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import numpy as np
from sparsegrad import *

from sparsegrad.forward import nvalue


def isscalar(x):
    return not forward.nvalue(x).shape


def getitem(x, idx):
    if idx is None:
        return x
    if not isscalar(idx) and not len(idx):
        return np.zeros(0, dtype=nvalue(x).dtype)
    if isscalar(x):
        return x
    else:
        return x[idx]


sparsesum = sparsesum_bare


def custom_function(f, df):
    f = expr.wrapped_func(func.custom_func(f, df))

    def _(*args):
        return f(*args)
    return f


def branch(cond, iftrue, iffalse):
    if isscalar(cond):
        if cond:
            return iftrue(None)
        else:
            return iffalse(None)
    n = len(cond)
    r = np.arange(len(cond))
    ixtrue = r[cond]
    ixfalse = r[np.logical_not(cond)]
    vtrue = iftrue(ixtrue)
    vfalse = iffalse(ixfalse)
    if isscalar(vtrue):
        vtrue = np.ones_like(ixtrue) * vtrue
    if isscalar(vfalse):
        vfalse = np.ones_like(ixfalse) * vfalse
    return sparsesum_bare(n, [(ixtrue, vtrue), (ixfalse, vfalse)])



# -*- coding: utf-8; -*-
#
# oedes - organic electronic device simulator
# Copyright (C) 2017 Marek Zdzislaw Szymanski (marek@marekszymanski.com)
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License, version 3,
# as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

from .base import *
from oedes import functions


class DirichletBC(BoundaryEquation):

    def __init__(self, name):
        self.name = name

    def setUp(self, eq):
        self.bidx = eq.mesh.getBoundary(self.name)['bidx']
        BoundaryEquation.setUp(self, eq)

    def _residuals(self, eq, v, FdS, b_value):
        i = eq.mesh.boundary.idx[self.bidx]
        return v[i] - b_value


class AppliedVoltage(DirichletBC):

    def residuals(self, eq, v, FdS, vars):
        assert isinstance(eq, Poisson)
        params = vars['params']
        return self._residuals(eq, v, FdS, params[
                               '%s.voltage' % self.name] - params['%s.workfunction' % self.name])


class Zero(DirichletBC):

    def residuals(self, eq, v, FdS, vars):
        return self._residuals(eq, v, FdS, 0)


class DirichletFromParams(DirichletBC):

    def residuals(self, eq, v, FdS, vars):
        return self._residuals(eq, v, FdS, vars['params'][
                               '%s.%s' % (eq.prefix, self.name)])


class Internal(BoundaryEquation):
    eq_from = None
    bidx_from = None

    def setUp(self, eq):
        BoundaryEquation.setUp(self, eq)
        self.conservation_to = self.eq_from.idx[
            self.eq_from.mesh.boundary.idx[self.bidx_from]]


class Equal(Internal):

    def __init__(self, other_eq, name):
        self.name = name
        self.eq_from = other_eq

    def setUp(self, eq):
        if self.eq_from is None:
            self.eq_from = eq
        bto, bfrom = eq.mesh.getSharedBoundary(self.eq_from.mesh, self.name)
        self.bidx = bto['bidx']
        self.bidx_from = bfrom['bidx']
        Internal.setUp(self, eq)

    def residuals(self, eq, v, FdS, vars):
        x = vars['x']
        return x[eq.idx[eq.mesh.boundary.idx[self.bidx]]] - \
            x[self.eq_from.idx[self.eq_from.mesh.boundary.idx[self.bidx_from]]]


class FermiLevelEqual(Equal):

    def residuals(self, eq, v, FdS, vars):
        ixto = eq.mesh.boundary.idx[self.bidx]
        ixfrom = self.eq_from.mesh.boundary.idx[self.bidx_from]
        Effrom = vars['eqvars'][id(self.eq_from)]['Ef'][
            self.eq_from.prefix][ixfrom]
        cto = vars['eqvars'][id(eq)]['model'].species_dos[
            self.eq_from.prefix].concentration(eq, vars, ixto, Effrom)
        return vars['x'][eq.idx[ixto]] - cto


class FermiLevelEqualElectrode(DirichletBC):
    F_eps = 1e-10

    def __init__(self, name, image_force=False, **kwargs):
        super(FermiLevelEqualElectrode, self).__init__(name, **kwargs)
        self.image_force = image_force

    def image_correction(self, eq, vars, ixto):
        if not self.image_force:
            return 0.
        n = eq.mesh.boundaries[self.name]['normal']
        F = eq.z * eq.mesh.dotv(vars['Ecellv'][ixto], n)
        F = where(F > self.F_eps, F, self.F_eps)
        return -eq.z * \
            functions.EmtageODwyerBarrierLowering(
                F, getitem(vars['epsilon'], ixto))

    def residuals(self, eq, v, FdS, vars):
        dos = vars['model'].species_dos[eq.prefix]
        ixto = eq.mesh.boundary.idx[self.bidx]
        params = vars['params']
        Ef_electrode = -params['%s.voltage' %
                               self.name] + self.image_correction(eq, vars, ixto)
        c_electrode = dos.concentration(eq, vars, ixto, Ef_electrode)
        return vars['x'][eq.idx[ixto]] - c_electrode



# -*- coding: utf-8; -*-
#
# oedes - organic electronic device simulator
# Copyright (C) 2017 Marek Zdzislaw Szymanski (marek@marekszymanski.com)
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License, version 3,
# as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

from oedes import model
from oedes.ad import *


class RestrictedModel(model):
    def __init__(self, model, eqs, x0, xt0):
        self.model = model
        self.idx = np.unique(np.hstack([model.findeq(n).idx for n in eqs]))
        self.bidx = np.zeros_like(model.X, dtype=np.bool)
        self.bidx[self.idx] = True
        self.nidx = np.arange(len(self.model.X))[np.logical_not(self.bidx)]
        self.dst = np.zeros_like(self.model.X, dtype=np.int)
        self.dst[self.idx] = np.arange(len(self.idx))
        self.X = np.zeros_like(self.idx, dtype=model.X.dtype)
        self.x0 = x0
        self.xt0 = xt0
        self.transientvar = model.transientvar[self.idx]
        self.eqs = eqs

    def _x_to_model(self, x, x0):
        return sparsesum(len(self.model.X), [
                         (self.idx, x), (self.nidx, x0[self.nidx])], unique=True, compress=False)

    def x_to_model(self, x):
        return self._x_to_model(x, self.x0)

    def evaluate(self, time, x, xt, params, **kwargs):
        model_X = self._x_to_model(x, self.x0)
        model_Xt = self._x_to_model(xt, self.xt0)
        for idx, f in self.model.evaluate(
                time, model_X, model_Xt, params, **kwargs):
            v = self.bidx[idx]
            yield self.dst[idx[v]], f[v]

    def update(self, x, params):
        model_X = self._x_to_model(x, self.x0)
        self.model.update(model_X, params)
        x[:] = model_X[self.idx]

    def scaling(self, params):
        xs, fs = self.model.scaling(params)
        return xs[self.idx], fs[self.idx]

    def converged(self, residuals, x, dx, params, report):
        model_residuals = self._x_to_model(residuals, 0. * self.model.X)
        model_x = self._x_to_model(x, 0. * self.model.X)
        model_dx = self._x_to_model(dx, 0. * self.model.X)
        report = {}
        r = self.model._converged(
            model_residuals,
            model_x,
            model_dx,
            params,
            report)
        return [report[k]['converged'] for k in self.eqs]
